package com.fg.scripting;

import com.fg.scripting.exception.*;

/**
 * This interface conceals logic connected with instantiating scripting (Groovy, JRuby ...) classes.
 * Each factory should contain constructor ScriptingFactory(ClassLoader parentClassLoader, Resource[] urls) in order
 * to by created by reflection in delegating proxies.
 *
 * @author Jan Novotný, FG Forrest a.s. (c) 2007
 * @version $Id: ScriptingFactory.java,v 1.5 2010/01/11 15:46:17 u_novoj Exp $
 */
@SuppressWarnings({"RawUseOfParameterizedType"})
public interface ScriptingFactory {

	/**
	 * Returns scripting classloader.
	 * @return this factory classloader
	 */
	ClassLoader getScriptingClassLoader();

	/**
	 * Returns true if class is loaded by ScriptingClassloader.
	 * @param clazz to analyze
	 * @return true if specified class is instantiated by this scripting factory classloader
	 */
	boolean isClassLoadedByScriptingClassloader(Class clazz);

	/**
	 * Loads class of specified name.
	 * @param className full class name with package
	 * @return class for specified name
	 * @throws ClassLoadingException whrown when source file cannot be found or is corrupted
	 */
	Class loadClass(String className) throws ClassLoadingException;

	/**
	 * Instantiates object of specified class.
	 * @param className full class name with package
	 * @param args constructor arguments
	 * @return class instance
	 * @throws ClassInstantiationException thrown when class instantiation fails
	 */
	Object createInstance(String className, Object... args) throws ClassInstantiationException;

	/**
	 * Instantiates object of specified class.
	 * @param clazz to be instantiated
	 * @param args constructor arguments
	 * @return class instance
	 * @throws ClassInstantiationException thrown when class instantiation fails
	 */
	Object createInstance(Class clazz, Object... args) throws ClassInstantiationException;

	/**
	 * Returns interval in miliseconds after that a class validity according to underlying source should be examined.
	 * @return refresh interval in miliseconds
	 */
	long getRefreshInterval();

	/**
	 * Closes ScriptingFactory, so no more queries will be handled by it.
	 */
	void close();

	/**
	 * Returns state of the scripting factory.
	 * @return true if factory is closed
	 */
	boolean isClosed();

}
