package com.fg.scripting.groovy;

import com.fg.scripting.*;
import com.fg.scripting.exception.*;
import com.fg.scripting.groovy.aop.*;
import com.fg.scripting.groovy.classloader.*;
import com.fg.scripting.groovy.monitoring.*;
import groovy.lang.GroovyClassLoader;
import groovy.util.GroovyScriptEngine;
import groovy.util.ResourceConnector;
import groovy.util.ResourceException;
import groovy.util.ScriptException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.groovy.control.CompilerConfiguration;
import org.springframework.aop.framework.ProxyFactory;
import org.springframework.aop.support.DefaultIntroductionAdvisor;
import org.springframework.aop.target.HotSwappableTargetSource;
import org.springframework.core.io.Resource;
import org.springframework.util.ClassUtils;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.net.URL;

/**
 * DevelopmentGroovyFactory creates instances of Groovy classes wrapped in proxies. This will allow us to safely hand
 * around references to those instances without worrying about behaviour refresh when underlying source code changes.
 * Proxies will maintain refreshing logic, checking source code change in specified intervals and eventually freeing
 * old instances and creating new fresh one based on new classes corresponding to latest source code state.
 *
 * Moreover the system is setup the way that, programmer doesn't need to take care of proper reference handling to avoid
 * PermGenSpace memory leaks. Returned references doesn't have anything common with Groovy class - they are derived from
 * first Java ancestor class and implement all interfaces loaded by pure Java classloader. All Groovy related things are
 * located in the proxy, so the swapping can be handled safely.
 *
 * @author Jan Novotný, FG Forrest a.s. (c) 2007
 * @version $Id: DevelopmentGroovyFactory.java,v 1.16 2010/05/20 12:11:38 u_novoj Exp $
 */
@SuppressWarnings({"RawUseOfParameterizedType"})
public class DevelopmentGroovyFactory implements ScriptingFactory {
	private static final Log log = LogFactory.getLog(DevelopmentGroovyFactory.class);
	private final GroovyScriptEngine engine;
	private final long refreshInterval;
	private ResourceConnector resourceConnector;
	private boolean closed;

	public DevelopmentGroovyFactory(ClassLoader parentClassLoader, Resource[] urls, String encoding) {
		this(parentClassLoader, new SimpleGroovyResourceLoader(urls), 0L, encoding);
	}

	public DevelopmentGroovyFactory(ClassLoader parentClassLoader, String[] urls, long refreshInterval, String encoding) {
		try {
			this.engine = new GroovyScriptEngine(urls, parentClassLoader);
			GroovyClassLoader classLoader = this.engine.getGroovyClassLoader();
			modifyConfigSourceEncoding(classLoader, encoding);
			this.refreshInterval = refreshInterval * 1000L;
			GroovyGarbageCollectorMonitor.addGroovyScriptEngineToMonitoring(engine);
		} catch(IOException ex) {
			String msg = "Invalid source urls: " + ex.getLocalizedMessage();
			log.error(msg);
			throw new RuntimeException(msg, ex);
		}
	}

	public DevelopmentGroovyFactory(ClassLoader parentClassLoader, URL[] urls, long refreshInterval, String encoding) {
		this.engine = new GroovyScriptEngine(urls, parentClassLoader);
		GroovyClassLoader classLoader = this.engine.getGroovyClassLoader();
		modifyConfigSourceEncoding(classLoader, encoding);
		this.refreshInterval = refreshInterval * 1000L;
		GroovyGarbageCollectorMonitor.addGroovyScriptEngineToMonitoring(engine);
	}

	public DevelopmentGroovyFactory(ClassLoader parentClassLoader, ResourceConnector rc, long refreshInterval, String encoding) {
		this.engine = new GroovyScriptEngine(rc, parentClassLoader);
		GroovyClassLoader classLoader = this.engine.getGroovyClassLoader();		
		modifyConfigSourceEncoding(classLoader, encoding);
		this.refreshInterval = refreshInterval * 1000L;
		this.resourceConnector = rc;
		GroovyGarbageCollectorMonitor.addGroovyScriptEngineToMonitoring(engine);
	}

	/**
	 * Sets encoding into the inner classloader - necessary hack as the config is by other mean unaccessible
	 * when using GroovyScriptEngine.
	 *
	 * TODO revisit when upgrading to 1.7.1
	 *
	 * @param classLoader
	 * @param encoding
	 */
	private void modifyConfigSourceEncoding(GroovyClassLoader classLoader, String encoding) {
		try {
			Field configField = classLoader.getClass().getSuperclass().getDeclaredField("config");
			configField.setAccessible(true);
			CompilerConfiguration config = (CompilerConfiguration)configField.get(classLoader);
			config.setSourceEncoding(encoding);
		} catch (Exception ex) {
			throw new IllegalStateException("Cannot set encoding on GroovyScriptingEngine.");
		}
	}

	/**
	 * Returns Groovy classloader.
	 *
	 * @return
	 */
	public ClassLoader getScriptingClassLoader() {
		checkClosed();
		return engine.getGroovyClassLoader();
	}

	/**
	 * Returns true if class is loaded by GroovyClassloader.
	 *
	 * @param clazz
	 * @return
	 */
	public boolean isClassLoadedByScriptingClassloader(Class clazz) {
		checkClosed();
		ClassLoader parameterClassLoader = clazz.getClassLoader();
		GroovyClassLoader engineClassLoader = engine.getGroovyClassLoader();
		do {
			if (parameterClassLoader == engineClassLoader) {
				return true;
			}
			parameterClassLoader = parameterClassLoader.getParent();
		} while (parameterClassLoader != ClassLoader.getSystemClassLoader());
		return false;
	}

	/**
	 * @see #createInstance(Class, Object...) 
	 *
	 * @param className
	 * @param args
	 * @return
	 * @throws ScriptException
	 * @throws ResourceException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	public Object createInstance(String className, Object... args) throws ClassInstantiationException {
		checkClosed();
		Class aClass = loadClass(className);
		return createInstance(aClass, args);
	}

	/**
	 * Creates groovy class instance. Returned instance is not directly groovy class instance but rather dynamic Proxy of
	 * that class. This proxy wraps original groovy class instance but extends not from this instance, but from nearest
	 * superclass that is loaded by Java classloader, also this proxy implements all interfaces, that are comming out of
	 * Java classloaders. Proxy itself encapsulates refreshing logic, so that if underlying instance doesn't conform to
	 * the source script it is automatically discarded and new instance is created instead. Reference to this proxy can
	 * be safely stored in long living memory scopes without having to fear that refreshing logic would cause pergenspace
	 * leaking. When you need to call method, that is defined only on the Groovy class but has no backing in Java superclass
	 * or any of interfaces, you can use the unwrap() method that returns reference of the wrapped Groovy instance. This
	 * reference must not be stored anywhere as it is a key to hell of permgenspace leaks.
	 *
	 * Single limitation of this approach is that you cannot change java superclass or java interface implementation set
	 * on the fly. Once generated proxy cannot adapt to this change.
	 *
	 * @param className
	 * @param args
	 * @return
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws ScriptException
	 * @throws ResourceException
	 */
	public Object createInstance(Class className, Object... args) throws ClassInstantiationException {
		checkClosed();
		Object instance;
		try {
			//load a groovy class and make an instance
			if (args.length == 0) {
				instance = className.newInstance();
			} else {
				Class[] parameterTypes = new Class[args.length];
				for(int i = 0; i < args.length; i++) {
					Object arg = args[i];
					parameterTypes[i] = arg != null ? arg.getClass() : null;
				}
				Constructor constructor = className.getConstructor(parameterTypes);
				instance = constructor.newInstance(args);
			}
		} catch (Exception ex) {
			String msg = "Cannot create instance of class: " + className.getName() + " (" + ex.getLocalizedMessage() + ')';
			log.error(msg, ex);
			throw new ClassInstantiationException(msg, ex, className);
		}

		if (isClassLoadedByScriptingClassloader(className)) {
			//analyze class and create proxy
			BuildProxyDescriptor bpd = getJavaAncestorsForProxy(className, engine.getParentClassLoader());
			ProxyFactory factory = new ProxyFactory(new Class[0]);
			factory.setTargetSource(
					new JavaRealmHotswappableTargetSource(instance, bpd.getSuperClassToExtend())
			);
			factory.setInterfaces(bpd.getInterfacesToImplement());
			factory.addAdvisor(new DefaultIntroductionAdvisor(new GroovyProxyMixin(this, instance)));
			factory.setProxyTargetClass(true);
			factory.setExposeProxy(true);
			factory.setFrozen(true);

			return factory.getProxy(this.getClass().getClassLoader());
		} else {
			return instance;
		}
	}

	/**
	 * Returns refresh interval passed to this object via constructor.
	 * @return
	 */
	public long getRefreshInterval() {
		checkClosed();
		return refreshInterval;
	}

	/**
	 * Loads class from GroovyScriptingEngine - respecting code source changes of this and dependent classes.
	 * Might return each time different class object instance - due to source code modifications.
	 * @param className
	 * @return
	 * @throws ScriptException
	 * @throws ResourceException
	 */
	public Class loadClass(String className) throws ClassLoadingException {
		checkClosed();
		try {
			//firstly imitate GroovyScriptingEngine and find out whether there is script of this name
			String scriptName = className.replace('.', File.separatorChar) + ".groovy";
			try {
				if (resourceConnector != null) {
					resourceConnector.getResourceConnection(scriptName);
				} else {
					engine.getResourceConnection(scriptName);
				}
				Class groovyClass = engine.loadScriptByName(className);
				//Groovy 1.7.0
				//Class groovyClass = engine.loadScriptByName(scriptName);
				GroovyGarbageCollectorMonitor.addGroovyClassToMonitoring(groovyClass);
				return groovyClass;
			} catch(ResourceException ignored) {
				//ups no script of this name found - lets load class on parent classloader
				return engine.getParentClassLoader().loadClass(className);
			}
		} catch (Exception ex) {
			String msg = "Cannot load class: " + className + " (" + ex.getLocalizedMessage() + ')';
			log.debug(msg, ex);
			throw new ClassLoadingException(msg, ex);
		}
	}

	/**
	 * Closes GroovyFactory, so no more queries will be handled by it.
	 */
	public void close() {
		this.closed = true;
	}

	/**
	 * Returns true if factory is closed.
	 *
	 * @return
	 */
	public boolean isClosed() {
		return closed;
	}

	/**
	 * Checks whether this GroovyFactory is closed and when so - IllegalStateException is thrown.
	 * @throws IllegalStateException
	 */
	private void checkClosed() throws IllegalStateException {
		if (closed) {
			throw new IllegalStateException("This GroovyFactory is closed - no more operations will be accepted.");
		}
	}

	/**
	 * Finds first superclass, that is loaded by Java classloader and all interfaces groovy class implements, that are
	 * also loaded by Java classloader.
	 * 
	 * @param groovyClass
	 * @param javaClassLoader
	 * @return
	 */
	private BuildProxyDescriptor getJavaAncestorsForProxy(Class groovyClass, ClassLoader javaClassLoader) {
		//find first ancestor that is loaded by pure Java classloader
		Class superClassToExtend = null;
		Class currentClass = groovyClass;
		do {
			currentClass = currentClass.getSuperclass();
			if (ClassUtils.isVisible(currentClass, javaClassLoader)) {
				superClassToExtend = currentClass;
			}
		} while (currentClass != null && superClassToExtend == null);
		//gather all interfaces, that are not loaded by Groovy classloader
		Class[] interfaces = ClassUtils.getAllInterfacesForClass(groovyClass, javaClassLoader);
		Class[] finalInterfaces = new Class[interfaces.length + 1];
		finalInterfaces[0] = ScriptingProxy.class;
		System.arraycopy(interfaces, 0, finalInterfaces, 1, interfaces.length);
		return new BuildProxyDescriptor(superClassToExtend, finalInterfaces);
	}

	/**
	 * Contains Java class description needed for Groovy proxy creation.
	 */
	public static class BuildProxyDescriptor {
		private final Class superClassToExtend;
		private final Class[] interfacesToImplement;

		public BuildProxyDescriptor(Class superClassToExtend, Class[] interfacesToImplement) {
			this.superClassToExtend = superClassToExtend;
			this.interfacesToImplement = interfacesToImplement;
		}

		public Class getSuperClassToExtend() {
			return superClassToExtend;
		}

		public Class[] getInterfacesToImplement() {
			return interfacesToImplement;
		}
	}

	/**
	 * Custom implementation of HotSwappableTargetSource that returns first Java parent class available for inner
	 * Groovy class instead of returning Groovy class itself.
	 */
	public static class JavaRealmHotswappableTargetSource extends HotSwappableTargetSource {
		private static final long serialVersionUID = -6962339257859210715L;
		private final Class superClassToExtend;

		public JavaRealmHotswappableTargetSource(Object target, Class superClassToExtend) {
			super(target);
			this.superClassToExtend = superClassToExtend;
		}

		@Override
		public synchronized Class getTargetClass() {
			return superClassToExtend;
		}

	}

}
