package com.fg.scripting.groovy;

import com.fg.scripting.*;
import com.fg.scripting.exception.*;
import com.fg.scripting.groovy.classloader.*;
import com.fg.scripting.groovy.monitoring.*;
import groovy.lang.GroovyClassLoader;
import groovy.lang.GroovyResourceLoader;
import groovy.util.ResourceException;
import groovy.util.ScriptException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.groovy.control.CompilerConfiguration;
import org.springframework.core.io.Resource;

import java.lang.reflect.Constructor;

/**
 * ProductionGroovyFactory creates instances of Groovy classes that never refresh. Returned instances are pure instances
 * of specified Groovy classes. References passed around will potentialy prevent Groovy class and classloader being
 * garbage collected, but that wouldn't be likely necessary in production (no refreshes occur there). On the other hand
 * this implementation represents shortest and easiest way to Groovy instances.
 *
 * @author Jan Novotný, FG Forrest a.s. (c) 2007
 * @version $Id: ProductionGroovyFactory.java,v 1.7 2010/05/20 12:11:38 u_novoj Exp $
 */
@SuppressWarnings({"ClassLoaderInstantiation"})
public class ProductionGroovyFactory implements ScriptingFactory {
	private static final Log log = LogFactory.getLog(ProductionGroovyFactory.class);
	private final GroovyClassLoader clsLoader;
	private boolean closed;

	public ProductionGroovyFactory(ClassLoader parentClassLoader, Resource[] baseDirs, String encoding) {
		this(parentClassLoader, new SimpleGroovyResourceLoader(baseDirs), encoding);
	}

	public ProductionGroovyFactory(ClassLoader parentClassLoader, SimpleGroovyResourceLoader rl, String encoding) {
		CompilerConfiguration configuration = new CompilerConfiguration();
		configuration.setSourceEncoding(encoding);
		this.clsLoader = new GroovyClassLoader(
				parentClassLoader,
				configuration
		);
		this.clsLoader.setShouldRecompile(false);
		this.clsLoader.setResourceLoader(rl);
		GroovyGarbageCollectorMonitor.addGroovyClassLoaderToMonitoring(clsLoader);
	}

	public ProductionGroovyFactory(ClassLoader parentClassLoader, GroovyResourceLoader rl, String encoding) {
		CompilerConfiguration configuration = new CompilerConfiguration();
		configuration.setSourceEncoding(encoding);
		this.clsLoader = new GroovyClassLoader(
				parentClassLoader,
				configuration
		);
		this.clsLoader.setShouldRecompile(false);
		this.clsLoader.setResourceLoader(rl);
		GroovyGarbageCollectorMonitor.addGroovyClassLoaderToMonitoring(clsLoader);
	}

	/**
	 * Returns Groovy classloader.
	 *
	 * @return
	 */
	public ClassLoader getScriptingClassLoader() {
		checkClosed();
		return clsLoader;
	}

	/**
	 * Returns true if class is loaded by GroovyClassloader.
	 *
	 * @param clazz
	 * @return
	 */
	public boolean isClassLoadedByScriptingClassloader(Class clazz) {
		checkClosed();
		ClassLoader parameterClassLoader = clazz.getClassLoader();
		do {
			if (parameterClassLoader == clsLoader) {
				return true;
			}
			parameterClassLoader = parameterClassLoader.getParent();
		} while (parameterClassLoader != ClassLoader.getSystemClassLoader());
		return false;
	}

	/**
	 * Loads Groovy class from GroovyClassLoader. The same className will always return the same class instance.
	 * @param className
	 * @return
	 * @throws ScriptException
	 * @throws ResourceException
	 * @throws ClassNotFoundException
	 */
	public Class loadClass(String className) throws ClassLoadingException {
		checkClosed();
		try {
			Class groovyClass = clsLoader.loadClass(className, true, true);
			GroovyGarbageCollectorMonitor.addGroovyClassToMonitoring(groovyClass);
			return groovyClass;
		} catch (Exception ex) {
			String msg = "Cannot load class: " + className + " (" + ex.getLocalizedMessage() + ")";
			log.error(msg, ex);
			throw new ClassLoadingException(msg, ex);
		}
	}

	/**
	 * @see #createInstance(Class, Object...)
	 *
	 * @param className
	 * @return
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws ScriptException
	 * @throws ResourceException
	 */
	public Object createInstance(String className, Object... args) throws ClassInstantiationException {
		checkClosed();
		Class aClass = loadClass(className);
		return createInstance(aClass, args);
	}

	/**
	 * Returns directly groovy instance with no modifications around.
	 *
	 * @param className
	 * @param args
	 * @return
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws ScriptException
	 * @throws ResourceException
	 */
	public Object createInstance(Class className, Object... args) throws ClassInstantiationException {
		checkClosed();
		Object groovyInstance;
		try {
			if (args.length == 0) {
				groovyInstance = className.newInstance();
			} else {
				Class[] parameterTypes = new Class[args.length];
				for(int i = 0; i < args.length; i++) {
					Object arg = args[i];
					parameterTypes[i] = arg != null ? arg.getClass() : null;
				}
				Constructor constructor = className.getConstructor(parameterTypes);
				groovyInstance = constructor.newInstance(args);
			}
		} catch (Exception ex) {
			String msg = "Cannot create instance of class: " + className.getName() + " (" + ex.getLocalizedMessage() + ")";
			log.error(msg, ex);
			throw new ClassInstantiationException(msg, ex, className);
		}
		return groovyInstance;
	}

	/**
	 * Returns maximum possible refresh interval - means no refresh will ever occur.
	 * @return
	 */
	public long getRefreshInterval() {
		checkClosed();
		return Long.MAX_VALUE;
	}

	/**
	 * Closes GroovyFactory, so no more queries will be handled by it.
	 */
	public void close() {
		this.closed = true;
	}

	/**
	 * Returns true if factory is closed.
	 *
	 * @return
	 */
	public boolean isClosed() {
		return closed;
	}

	/**
	 * Checks whether this GroovyFactory is closed and when so - IllegalStateException is thrown.
	 * @throws IllegalStateException
	 */
	private void checkClosed() throws IllegalStateException {
		if (closed) {
			throw new IllegalStateException("This GroovyFactory is closed - no more operations will be accepted.");
		}
	}

}
