package com.fg.scripting.web;

import com.fg.scripting.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.core.io.Resource;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.ServletContextResource;
import org.springframework.web.context.support.WebApplicationContextUtils;

import javax.servlet.ServletContext;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;

/**
 * Shared utils for servlet proxies.
 *
 * @author Jan Novotný, FG Forrest a.s. (c) 2007
 * @version $Id: ScriptingDelegationUtils.java,v 1.2 2010/01/14 12:30:27 u_novoj Exp $
 */
public class ScriptingDelegationUtils {
	private static final Log log = LogFactory.getLog(ScriptingDelegationUtils.class);

	/**
	 * Returns true if string is empty or null.
	 * @param targetClass
	 * @return
	 */
	public static boolean isEmpty(String targetClass) {
		return targetClass == null || targetClass.trim().length() == 0;
	}

	/**
	 * Creates new ScriptingFactory from the specified parameters. Handles all error checking and logging.
	 *
	 * @param factoryClassName default factoryClass when parameter factoryClassName is null is ProductionGroovyFactory.
	 * @param scriptDirectory required
	 * @param servletContext reqiuired
	 * @return if anything fails, error is logged and null is returned back, otherwise initialized ScriptingFactory
	 */
	public static ScriptingFactory getNewScriptingFactory(String factoryClassName, String scriptDirectory, ServletContext servletContext, String encoding) {
		if (isEmpty(factoryClassName)) {
			factoryClassName = "com.fg.scripting.groovy.ProductionGroovyFactory";
		}

		Class<?> factoryClass;
		try {
			factoryClass = Class.forName(factoryClassName);
			if (!ScriptingFactory.class.isAssignableFrom(factoryClass)) {
				String msg = "Factory class " + factoryClassName + " (factory init param) doesn't represent valid ScriptingFactory implementation!";
				log.error(msg);
				return null;
			}
		}
		catch(ClassNotFoundException e) {
			String msg = "ScriptingFactory " + factoryClassName + " not found on classpath: " + e.getLocalizedMessage();
			log.error(msg, e);
			return null;
		}

		Constructor<?> reflexConstructor;
		try {
			reflexConstructor = factoryClass.getDeclaredConstructor(ClassLoader.class, Resource[].class, String.class);
		}
		catch(NoSuchMethodException e) {
			String msg = "No constructor with signature ScriptingFactory(ClassLoader parentClassLoader, Resource[] urls) found in class: " + factoryClassName;
			log.error(msg, e);
			return null;
		}

		if (isEmpty(scriptDirectory)) {
			String msg = "No directories with scripting sources specified no 'scriptDir' init parameter defined or is empty!";
			log.error(msg);
			return null;
		}

		WebApplicationContext springContext = WebApplicationContextUtils.getWebApplicationContext(servletContext);
		String[] directories = scriptDirectory.split("\n");
		List<Resource> baseResources = new ArrayList<Resource>();
		for(String targetPath : directories) {
			if(!isEmpty(targetPath)) {
				baseResources.add(
					springContext != null ?
							springContext.getResource(targetPath) :
							new ServletContextResource(servletContext, targetPath)
				);
			}
		}

		try {
			return (ScriptingFactory)reflexConstructor.newInstance(
					Thread.currentThread().getContextClassLoader(),
					baseResources.toArray(new Resource[baseResources.size()]),
					encoding
			);
		}
		catch(Exception e) {
			String msg = "Cannot instantiate scripting factory " + factoryClassName + " due to:" + e.getLocalizedMessage();
			log.error(msg, e);
			return null;
		}
	}
}
