package com.fg.scripting.groovy;

import groovy.lang.GroovyClassLoader;
import groovy.util.GroovyScriptEngine;
import groovy.util.ResourceException;
import groovy.util.ScriptException;
import junit.framework.TestCase;
import org.apache.commons.io.IOUtils;
import org.springframework.core.io.FileSystemResource;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * Description
 *
 * @author Jan Novotný, FG Forrest a.s. (c) 2007
 * @version $Id: AbstractGroovyTest.java,v 1.6 2010/01/15 14:40:49 u_novoj Exp $
 */
public abstract class AbstractGroovyTest extends TestCase {
	protected FileSystemResource rootFSR;
	private static final String SOURCE_ENCODING = "windows-1250";
	private static final String BACKUP_EXTENSION = ".backup";

	@Override
	protected void setUp() throws Exception {
		super.setUp();
		//support for TeamCity runs
		String baseDir = "/www/p_java/cps/libs_java/lib_groovy";
		if(System.getProperty("teamcity.build.checkoutDir") != null) {
			baseDir = System.getProperty("teamcity.build.checkoutDir");
		}
		System.setProperty("project.build.directory", baseDir);
		rootFSR = new FileSystemResource(baseDir + "/target/test-classes/");
		//backup test groovy classes
		PathMatchingResourcePatternResolver resolver = new PathMatchingResourcePatternResolver();
		Resource[] resources = resolver.getResources("**/*.groovy");
		for(Resource resource : resources) {
			Reader reader = null;
			Writer writer = null;
			try {
				reader = new InputStreamReader(new FileInputStream(resource.getFile()), SOURCE_ENCODING);
				File newFile = new File(resource.getFile().getAbsolutePath() + BACKUP_EXTENSION);
				writer = new OutputStreamWriter(new FileOutputStream(newFile), SOURCE_ENCODING);
				IOUtils.copy(reader, writer);
			}
			finally {
				IOUtils.closeQuietly(reader);
				IOUtils.closeQuietly(writer);
			}
		}
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
		//restore original groovy classes
		PathMatchingResourcePatternResolver resolver = new PathMatchingResourcePatternResolver();
		Resource[] resources = resolver.getResources("**/*.groovy" + BACKUP_EXTENSION);
		for(Resource resource : resources) {
			Reader reader = null;
			Writer writer = null;
			try {
				reader = new InputStreamReader(new FileInputStream(resource.getFile()), SOURCE_ENCODING);
				String resourceAbsolutePath = resource.getFile().getAbsolutePath();
				File newFile = new File(resourceAbsolutePath.substring(0, resourceAbsolutePath.length() - BACKUP_EXTENSION.length()));
				writer = new OutputStreamWriter(new FileOutputStream(newFile), SOURCE_ENCODING);
				IOUtils.copy(reader, writer);
			}
			finally {
				IOUtils.closeQuietly(reader);
				IOUtils.closeQuietly(writer);
			}
		}
	}

	protected FileSystemResource modifyGroovyFile(String groovyFile, String replaceWhat, String replaceWith) throws IOException, InterruptedException {
		//wait a second to ensure file gets new timestamp
		synchronized(this) {
			this.wait(1000);
		}
		FileSystemResource resource = (FileSystemResource)rootFSR.createRelative(groovyFile);
		String scriptContents;
		InputStream is = resource.getInputStream();
		try {
			scriptContents = IOUtils.toString(is, "UTF-8");
		} finally {
			IOUtils.closeQuietly(is);
		}
		String modifiedScriptContents = scriptContents.replaceAll(replaceWhat, replaceWith);
		OutputStream output = new FileOutputStream(resource.getFile());
		try {
			IOUtils.write(modifiedScriptContents, output);
		} finally {
			IOUtils.closeQuietly(output);
		}
		return resource;
	}

	protected GroovyContext createInstanceAndCallMethod(GroovyScriptEngine engine, String className, String methodName, Object... args) throws ResourceException, ScriptException, InstantiationException, IllegalAccessException, NoSuchMethodException, InvocationTargetException {
		Class MainGroovyClass = engine.loadScriptByName(className);
		//needed from Groovy 1.7.0
		//Class MainGroovyClass = engine.loadScriptByName(className.replace('.', File.separatorChar) + ".groovy");
		Object myGroovy = MainGroovyClass.newInstance();
		Class[] params = new Class[args.length];
		for(int i = 0; i < args.length; i++) {
			Object arg = args[i];
			params[i] = arg.getClass();
		}
		Method helloWorldMethod = MainGroovyClass.getMethod(methodName, params);
		Object result = helloWorldMethod.invoke(myGroovy, args);
		return new GroovyContext(MainGroovyClass, myGroovy, helloWorldMethod, result);
	}

	protected GroovyContext createInstanceAndCallMethod(GroovyClassLoader clsLoader, String className, String methodName, Object... args) throws ResourceException, ScriptException, InstantiationException, IllegalAccessException, NoSuchMethodException, InvocationTargetException, ClassNotFoundException {
		Class mainGroovyClass = clsLoader.loadClass(className, true, false);
		Object myGroovy = mainGroovyClass.newInstance();
		Class[] params = new Class[args.length];
		for(int i = 0; i < args.length; i++) {
			Object arg = args[i];
			params[i] = arg.getClass();
		}
		Method helloWorldMethod = mainGroovyClass.getMethod(methodName, params);
		Object result = helloWorldMethod.invoke(myGroovy, args);
		return new GroovyContext(mainGroovyClass, myGroovy, helloWorldMethod, result);
	}

	protected Object callMethod(Object instance, String methodName, Object... args) throws Exception {
		Class[] params = new Class[args.length];
		for(int i = 0; i < args.length; i++) {
			Object arg = args[i];
			params[i] = arg.getClass();
		}
		Method helloWorldMethod = instance.getClass().getMethod(methodName, params);
		return helloWorldMethod.invoke(instance, args);
	}

	protected static class GroovyContext {
		private final Class groovyClass;
		private final Object groovyInstance;
		private final Method groovyMethod;
		private final Object groovyResult;

		private GroovyContext(Class groovyClass, Object groovyInstance, Method groovyMethod, Object groovyResult) {
			this.groovyClass = groovyClass;
			this.groovyInstance = groovyInstance;
			this.groovyMethod = groovyMethod;
			this.groovyResult = groovyResult;
		}

		public Class getGroovyClass() {
			return groovyClass;
		}

		public Object getGroovyInstance() {
			return groovyInstance;
		}

		public Method getGroovyMethod() {
			return groovyMethod;
		}

		public Object getGroovyResult() {
			return groovyResult;
		}
	}
}
